module Formula (Formula(..), tosmt, formulaChangeCore, (//), getCoreList) where
import Debug.Trace (trace)
import Data.Traversable (Traversable(), traverse)
import Data.Foldable (Foldable(), foldr, toList)
import Control.Applicative ((<$>),(<*>))
import Prelude hiding (traverse, foldr)

instance Functor Formula where
  fmap f = formulaChangeCore (BASIC . f)

instance Foldable Formula where
  foldr f i (BASIC a)    = f a i
  foldr f i (AND lst)    = foldr (\x y -> foldr f y x) i lst
  foldr f i (OR  lst)    = foldr (\x y -> foldr f y x) i lst
  foldr f i (EQUIV a b)  = foldr f (foldr f i b) a
  foldr f i (IMPL a b)   = foldr f (foldr f i b) a
  foldr f i (NOT  a)     = foldr f i a
  foldr f i (COMMENT _ a)= foldr f i a

instance Traversable Formula where
  traverse f (BASIC a)    = BASIC     <$> f a
  traverse f (AND lst)    = AND       <$> traverse (traverse f) lst
  traverse f (OR  lst)    = OR        <$> traverse (traverse f) lst
  traverse f (EQUIV a b)  = EQUIV     <$> (traverse f a) <*> (traverse f b)
  traverse f (IMPL a b)   = IMPL      <$> (traverse f a) <*> (traverse f b)
  traverse f (NOT  a)     = NOT       <$> (traverse f a)
  traverse f (COMMENT s a)= COMMENT s <$> (traverse f a)

(//) a s = COMMENT s a
infixr 4 //

formulaChangeCore :: (a -> Formula b) -> (Formula a) -> (Formula b)
formulaChangeCore f (AND lst)    = AND       (map (formulaChangeCore f) lst)
formulaChangeCore f (OR   lst)   = OR        (map (formulaChangeCore f) lst)
formulaChangeCore f (EQUIV a b)  = EQUIV     (formulaChangeCore f a) (formulaChangeCore f b)
formulaChangeCore f (IMPL a b)   = IMPL      (formulaChangeCore f a) (formulaChangeCore f b)
formulaChangeCore f (NOT  a)     = NOT       (formulaChangeCore f a)
formulaChangeCore f (COMMENT s a)= COMMENT s (formulaChangeCore f a)
formulaChangeCore f (BASIC a)    = f a

getCoreList :: (Formula a) -> [a]
getCoreList = toList

-- some legacy code may have: AND String [Formula a]
-- the error could be either:
 -- (1) AND is applied too many arguments
 -- (2) could not match `Char' to `Formula a'
-- this should be changed to: COMMENT String (AND [Formula a])
data Formula a = AND [Formula a] | OR [Formula a]
               | EQUIV (Formula a) (Formula a)
               | IMPL (Formula a) (Formula a)
               | NOT (Formula a)
               | COMMENT String (Formula a) -- add a comment to the formula. Careful though: rewriting will not interpret COMMENT!
               | BASIC a
tosmt :: Int -> (a->String) -> (Formula a) -> String
tosmt a b  (AND [c]) = tosmt a b c
tosmt a b  (OR  [c]) = tosmt a b c
tosmt _ _  (AND [] ) = "true"
tosmt _ _  (OR  [] ) = "false"
tosmt _ ds (BASIC a) = ds a
tosmt i ds (COMMENT s a) = "; " ++ s ++ "\n"++ take i (repeat ' ') ++ tosmt i ds a
tosmt i ds x = "("++tosmt' x++")"
 where idt = "\n"++take i (repeat ' ')
       m op a b = op ++ idt ++ (take (length op + 2) (repeat ' '))
                  ++ tosmt (i+length op+2) ds a ++ idt ++ (take (length op + 2) (repeat ' '))
                  ++ tosmt (i+length op+2) ds b
       tosmt' (IMPL a b)     = m "=>" a b
       tosmt' (EQUIV a b)    = m "=" a b
       tosmt' (AND as)       = "and " ++ concat [idt ++ "  "++ tosmt (i+2) ds a | a <- as]
       tosmt' (OR  as)       = "or" ++ concat [idt ++ "  "++ tosmt (i+2) ds a | a <- as]
       tosmt' (NOT a)        = "not " ++ tosmt (i+5) ds a
       tosmt' (BASIC a)      = ds a
       tosmt' (COMMENT s a)  = trace "Created ;;; in the output, please fix!"$ ";;;; " ++ s ++ idt ++ tosmt' a