// Queue of depth 2
`include "global.vh"
module xmas_queue_2 (clk, rst,
    i$irdy, i$trdy, i$data,
    o$irdy, o$trdy, o$data);

input clk, rst;
    
input i$irdy;
output i$trdy;
input [`DATA_WIDTH-1:0] i$data;

output o$irdy;
input o$trdy;
output [`DATA_WIDTH-1:0] o$data;

// buffer 0 is the header. Packets enter at the lowest empty buffer, and move towards the header.

reg [0:0] in;
reg [0:0] out;
wire [0:0] nextin = (in==2-1)?0:(in+1);
wire [0:0] nextout = (out==2-1)?0:(out+1);
// Queue memory
reg [`DATA_WIDTH-1:0] data0;
reg [`DATA_WIDTH-1:0] data1;
reg full;

assign i$trdy = !full;
assign o$irdy = !(in==out) || full;

assign o$data = out==0?data0:(out==1?data1:(data1));
wire writing = i$irdy && i$trdy; // writing into the queue
wire reading = o$irdy && o$trdy; // reading from the queue

always @(posedge clk) in   <= rst ? 0 : (writing ? nextin : in);
always @(posedge clk) out  <= rst ? 0 : (reading ? nextout : out);
always @(posedge clk) full <= (rst || reading) ? 0 : ((nextin==out && writing) ? 1 : full);

always @(posedge clk) data0 <= (writing && (in==0)) ? i$data : data0;
always @(posedge clk) data1 <= (writing && (in==1)) ? i$data : data1;
endmodule
